from .model import *
from .audio_classification import *
from .image2text import *
from .language_modeling import *
from .sequence_labeling import *
from .speech_recognition import *
from .text_classification import *
from .text_detection import *
from .text_summarization import *



from re import sub
from typing import Dict, List


__all__ = [
    "snake_case",
    "permute_dict_list",
]


def snake_case(s):
    return '_'.join(
        sub('([A-Z][a-z]+)', r' \1',
            sub('([A-Z]+)', r' \1',
                s.replace('-', ' '))).split()).lower()


def permute_dict_list(dict_list: List[Dict]) -> Dict[str, List]:
    """
    Convert a list of dictionaries to a dictionary of lists

    Args:
        dict_list: Input list of dicts

    Returns:

    """
    if not len(dict_list):
        return {}
    d = {key: [x[key] for x in dict_list] for key in dict_list[0]}
    return d



import numpy as np
import torch

from ...configs import TrainerConfig
from ...constants import MetricType
from ...data.datasets import Dataset
from ...models import Model
from ...utils import get_logger
from ..trainer import Trainer


logger = get_logger(__name__)


class SequenceLabelingTrainer(Trainer):
    """
    A trainer for all text classification models

    Args:
        model ([`Model`] or `torch.nn.Module`): The main model to train and evaluate
        config (TrainerConfig): Training configuration and parameters
        train_dataset (Dataset): Train dataset
        eval_dataset (Dataset): Evaluation dataset
        data_collator: Collate function, usually included in the dataset object itself
        optimizer (optim.Optimizer): Model optimizer
        lr_scheduler: Optional learning-rate scheduler

    """
    AVAILABLE_METRICS = [MetricType.SEQEVAL]

    def __init__(
        self,
        model: Model = None,
        config: TrainerConfig = None,
        train_dataset: Dataset = None,
        eval_dataset: Dataset = None,
        data_collator=None,
        optimizer: torch.optim.Optimizer = None,
        lr_scheduler=None,
    ):
        super().__init__(
            model=model,
            config=config,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            data_collator=data_collator,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
        )
        self.criterion = torch.nn.CrossEntropyLoss()

    def compute_loss(self, logits, labels, **kwargs) -> torch.Tensor:
        loss = self.criterion(logits.view(-1, self.model.config.num_labels), labels.view(-1))
        return loss

    def compute_metrics(self, predictions, labels, **kwargs):
        predictions = np.array(predictions).argmax(2).squeeze()
        labels = np.array(labels).squeeze()

        # Remove ignored index (special tokens) and append `B-` in the beginning for seqeval
        true_predictions = [
            [f"B-{self.model.config.id2label[p]}" for (p, l) in zip(prediction, label) if l != -100]
            for prediction, label in zip(predictions, labels)
        ]
        true_labels = [
            [f"B-{self.model.config.id2label[l]}" for (p, l) in zip(prediction, label) if l != -100]
            for prediction, label in zip(predictions, labels)
        ]

        results = {}
        for metric_name, metric in self.metrics.items():
            results[metric_name] = metric.compute(true_predictions, true_labels)["f1"]
        return results




import logging


__all__ = [
    "get_logger",
]


def get_logger(name, level=None, fmt=None):
    fmt = fmt or "%(levelname)s: %(message)s"
    level = level or "INFO"

    logger = logging.Logger(name, level)
    handler = logging.StreamHandler()
    formatter = logging.Formatter(fmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger



from dataclasses import dataclass
from typing import Iterable

from sklearn.metrics import f1_score

from ..configs import MetricConfig
from ..constants import MetricType
from ..registry import register_metric
from .metric import Metric


@dataclass
class F1Config(MetricConfig):
    name: str = MetricType.F1
    pos_label: int = None
    average: str = None
    sample_weight: Iterable[float] = None


@register_metric("f1", config_class=F1Config)
class F1(Metric):
    def __init__(self, config: F1Config, **kwargs):
        super().__init__(config, **kwargs)

    def compute(
        self,
        predictions=None,
        targets=None,
        labels=None,
        pos_label=1,
        average=None,
        sample_weight=None
    ):
        pos_label = pos_label or self.config.pos_label
        average = average or self.config.average
        sample_weight = sample_weight or self.config.sample_weight

        score = f1_score(
            targets,
            predictions,
            labels=labels,
            pos_label=pos_label,
            average=average,
            sample_weight=sample_weight,
        )

        return float(score) if score.size == 1 else score



import os
import random
from typing import Any, Dict, Tuple

import numpy as np
import torch
from huggingface_hub import create_repo, hf_hub_download
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from ..builders import build_metric
from ..configs import LRSchedulerConfig, MetricConfig, OptimizerConfig, TrainerConfig
from ..constants import (
    DEFAULT_DATASET_CONFIG_FILE,
    DEFAULT_TRAINER_CONFIG_FILE,
    DEFAULT_TRAINER_SUBFOLDER,
    HEZAR_CACHE_DIR,
    TQDM_BAR_FORMAT,
)
from ..data.datasets import Dataset
from ..models import Model
from ..utils import get_logger
from .trainer_utils import MetricsTracker


logger = get_logger(__name__)

optimizers = {
    "adam": torch.optim.Adam,
    "adamw": torch.optim.AdamW,
    "sgd": torch.optim.SGD,
}
lr_schedulers = {
    "reduce_on_plateau": torch.optim.lr_scheduler.ReduceLROnPlateau,
    "cosine_lr": torch.optim.lr_scheduler.CosineAnnealingLR,
}


class Trainer:
    """
    A general but fully featured model trainer/evaluator for Hezar models.

    Args:
        model ([`Model`] or `torch.nn.Module`): The main model to train and evaluate
        config (TrainerConfig): Training configuration and parameters
        train_dataset (Dataset): Train dataset
        eval_dataset (Dataset): Evaluation dataset
        data_collator: Collate function, usually included in the dataset object itself
        optimizer (optim.Optimizer): Model optimizer
        lr_scheduler: Optional learning-rate scheduler

    """

    trainer_subfolder = DEFAULT_TRAINER_SUBFOLDER
    trainer_config_file = DEFAULT_TRAINER_CONFIG_FILE
    dataset_config_file = DEFAULT_DATASET_CONFIG_FILE
    AVAILABLE_METRICS = []

    def __init__(
        self,
        model: Model = None,
        config: TrainerConfig = None,
        train_dataset: Dataset = None,
        eval_dataset: Dataset = None,
        data_collator=None,
        optimizer: torch.optim.Optimizer = None,
        lr_scheduler=None,
        compute_metrics=None,
        **kwargs,
    ):

        self.config = config

        self.device, self.device_type = self._prepare_device_and_type()
        self.autocast_dtype = torch.bfloat16 if self.device_type == "cpu" else torch.float16
        self.scaler = torch.cuda.amp.GradScaler(enabled=self.config.use_amp and self.device_type == "cuda")

        self._set_seed(self.config.seed)

        self.model = self._prepare_model(model)

        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset
        self.data_collator = data_collator or self.train_dataset.data_collator
        self.train_dataloader, self.eval_dataloader = self._prepare_dataloaders()

        self.optimizer, self.lr_scheduler = self._prepare_optimizers(optimizer, lr_scheduler)

        self.metrics = self.setup_metrics()
        self.metrics_tracker = MetricsTracker(list(self.metrics.keys()))

        self.tensorboard = SummaryWriter(log_dir=self.config.log_dir)

    def _prepare_device_and_type(self) -> Tuple[str, str]:
        device = self.config.device if "cuda" in self.config.device and torch.cuda.is_available() else "cpu"
        device_type = "cuda" if "cuda" in device else "cpu"
        return device, device_type

    @staticmethod
    def _set_seed(seed) -> None:
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)

    def _prepare_model(self, model: Model) -> Model:
        """
        Download the model from HuggingFace Hub if `init_weights_from` is given in the config. Load the model to the
        device and return it.
        :param model:
        :return:
        """
        if model is None:
            raise ValueError("`model` must be given to the Trainer!")
        hub_path = self.config.init_weights_from
        if hub_path is not None:
            local_path = hf_hub_download(hub_path, filename=model.model_filename, cache_dir=HEZAR_CACHE_DIR)
            model.load_state_dict(torch.load(local_path, map_location="cpu"))
        model.to(self.device)
        return model

    def _prepare_dataloaders(self):
        """
        Set up data loaders (train/eval) and return them.

        Returns:
             A tuple of train and eval dataloaders
        """
        if self.train_dataset is not None:
            train_dataloader = DataLoader(
                dataset=self.train_dataset,
                batch_size=self.config.batch_size,
                collate_fn=self.data_collator,
                num_workers=self.config.num_dataloader_workers,
                drop_last=True,
                shuffle=True,
            )
        else:
            raise ValueError("Cannot create train dataloader because `train_dataset` is not given!")
        if self.eval_dataset is not None:
            eval_dataloader = DataLoader(
                dataset=self.eval_dataset,
                batch_size=self.config.batch_size,
                collate_fn=self.data_collator,
                num_workers=self.config.num_dataloader_workers,
                drop_last=True,
                shuffle=True,
            )
        else:
            logger.warning("Cannot create eval dataloader because `eval_dataset` is not given to the Trainer!")
            eval_dataloader = None

        return train_dataloader, eval_dataloader

    def _prepare_optimizers(self, optimizer: torch.optim.Optimizer = None, lr_scheduler=None):
        """
        Set up the optimizer and lr scheduler if they're not already given

        Args:
            optimizer: If None do nothing and return it, otherwise build it using the train config
            lr_scheduler: If None do nothing and return it, otherwise build it using the train config

        Returns:
            Optimizer and scheduler
        """
        if optimizer is None:
            optimizer_config = self.config.optimizer

            # convert to dict so that we can pop some values
            if isinstance(optimizer_config, OptimizerConfig):
                optimizer_config = optimizer_config.dict()

            optimizer_name = optimizer_config.pop("name")
            scheduler_config = optimizer_config.pop("scheduler")

            optimizer_config.pop("config_type", None)
            optimizer = optimizers[optimizer_name](self.model.parameters(), **optimizer_config)

            if lr_scheduler is None and scheduler_config is not None:
                if isinstance(scheduler_config, LRSchedulerConfig):
                    scheduler_config = scheduler_config.dict()
                scheduler_name = scheduler_config.pop("name")
                scheduler_config.pop("config_type", None)
                lr_scheduler = lr_schedulers[scheduler_name](optimizer, **scheduler_config)
        return optimizer, lr_scheduler

    def setup_metrics(self):
        metrics_dict = {}
        for metric in self.config.metrics:
            if isinstance(metric, str):
                if metric not in self.AVAILABLE_METRICS:
                    raise ValueError(f"Invalid metric `{metric}`! Available metrics: {self.AVAILABLE_METRICS}")
                metrics_dict[metric] = build_metric(metric)
            elif isinstance(metric, MetricConfig):
                metrics_dict[metric.name] = build_metric(metric.name, config=metric)
            else:
                raise ValueError(f"Invalid metric type `{type(metric)}`! Available metrics: {self.AVAILABLE_METRICS}")
        return metrics_dict

    def prepare_input_batch(self, input_batch):
        """
        Every operation required to prepare the inputs for model forward like moving to device, permutations, etc.
        Args:
            input_batch: Raw input batch from the dataloader

        Returns:
            The proper input batch required by model forward
        """
        # cast to device
        input_batch = {k: v.to(self.device) for k, v in input_batch.items() if isinstance(v, torch.Tensor)}
        return input_batch

    def amp_context_manager(self):
        """
        A smart context manager for mixed precision.

        Returns:
            A torch autocast context manager
        """
        return torch.autocast(device_type=self.device_type, dtype=self.autocast_dtype, enabled=self.config.use_amp)

    def forward(self, input_batch):
        """
        Perform model forward on the input batch

        In special cases, one can override this method in their desired trainer.

        Args:
            input_batch: Input batch

        Returns:
            Model outputs
        """
        outputs = self.model(input_batch)
        return outputs

    def compute_loss(self, logits: torch.Tensor, labels: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Perform model forward and compute loss.

        This method must be implemented in other trainers.

        Args:
            logits: Logits from model outputs
            labels: Ground truth labels

        Returns:
            The loss tensor
        """
        raise NotImplementedError

    def compute_metrics(self, predictions, labels, **kwargs):
        """
        Compute metric values on the predictions and labels

        Args:
            predictions: A list of all predictions
            labels: A list of all labels

        Returns:
            A dictionary of the results for every metric specified by the trainer
        """
        return {}

    def training_step(self, input_batch: Dict[str, torch.Tensor]) -> Dict[str, Any]:
        """
        Train one batch of data and return loss and model outputs

        Args:
            input_batch: A batch of inputs to train

        Returns:
            Train step outputs including loss, logits, etc.
        """
        with self.amp_context_manager():
            outputs = self.forward(input_batch)
            loss = self.compute_loss(outputs["logits"], input_batch["labels"])

        self.scaler.scale(loss).backward()
        self.scaler.step(self.optimizer)
        self.scaler.update()
        self.optimizer.zero_grad()

        outputs["loss"] = loss.item() if isinstance(loss, torch.Tensor) else loss

        return outputs

    def evaluation_step(self, input_batch: Dict[str, torch.Tensor]) -> Dict[str, Any]:
        """
        Evaluate one batch of data and return loss and model outputs

        Args:
            input_batch: A batch of inputs to evaluate

        Returns:
            Evaluation step outputs including loss, logits, etc.
        """
        with self.amp_context_manager():
            outputs = self.forward(input_batch)
            loss = self.compute_loss(outputs["logits"], input_batch["labels"])

        outputs["loss"] = loss.item() if isinstance(loss, torch.Tensor) else loss

        return outputs

    def inner_training_loop(self, epoch_num: int):
        """
        Train the model for one epoch on the whole train dataset and verbose live metric values in the progress bar

        Args:
            epoch_num: Number of the current epoch

        Returns:
            Metrics averages through the full iteration
        """
        self.metrics_tracker.reset()
        self.model.train()
        with tqdm(
            self.train_dataloader,
            unit="batch",
            desc=f"Epoch: {epoch_num}/{self.config.num_epochs} ",
            bar_format=TQDM_BAR_FORMAT,
            ascii=" #",
        ) as iterator:
            for step, input_batch in enumerate(iterator):
                input_batch = self.prepare_input_batch(input_batch)
                # Training on one batch
                outputs = self.training_step(input_batch)
                # Compute metrics
                training_results = self.compute_metrics(
                    outputs["logits"].detach().cpu().numpy(),
                    input_batch["labels"].detach().cpu().numpy(),
                )
                training_results["loss"] = outputs["loss"]
                # Gather outputs for metrics
                self.metrics_tracker.update(training_results)
                iterator.set_postfix(**self.metrics_tracker.avg())

        return self.metrics_tracker.avg()

    def evaluate(self):
        """
        Evaluates the model on the whole eval dataset and verbose live metric values in the progress bar

        Returns:
            Evaluation results
        """
        self.metrics_tracker.reset()
        self.model.eval()
        with tqdm(
            self.eval_dataloader,
            unit="batch",
            desc="Evaluating... ",
            bar_format=TQDM_BAR_FORMAT,
            ascii=" #",
        ) as iterator:
            with torch.inference_mode():
                for step, input_batch in enumerate(iterator):
                    input_batch = self.prepare_input_batch(input_batch)
                    # Evaluation on one batch
                    outputs = self.evaluation_step(input_batch)
                    # Compute metrics
                    evaluation_results = self.compute_metrics(
                        outputs["logits"].detach().cpu().numpy(),
                        input_batch["labels"].detach().cpu().numpy(),
                    )
                    evaluation_results["loss"] = outputs["loss"]
                    # Gather outputs for metrics
                    self.metrics_tracker.update(evaluation_results)
                    iterator.set_postfix(**self.metrics_tracker.avg())

        return self.metrics_tracker.avg()

    def train(self):
        """
        The full training process like training, evaluation, logging and saving model checkpoints.
        """
        for epoch in range(1, self.config.num_epochs + 1):
            print()
            self.inner_training_loop(epoch)
            evaluation_results = self.evaluate()
            self.lr_scheduler.step(evaluation_results["loss"])

            # maybe save checkpoint
            if epoch % self.config.save_freq == 0:
                ckpt_save_path = os.path.join(self.config.checkpoints_dir, str(epoch))
                self.save(ckpt_save_path)

    def save(
        self,
        path: str,
        config_filename=None,
        model_filename=None,
        model_config_filename=None,
        subfolder=None,
        dataset_config_file=None,
    ):
        """
        Save the trainer and relevant files to a path.

        Files to save are train config, model weights, model config, preprocessor files and preprocessor config.

        Args:
            path: A directory to save everything
            config_filename: Config filename
            model_filename: Model file name
            model_config_filename: Model config file name
            subfolder: Optional sub-folder
            dataset_config_file: Dataset config filename
        """
        config_filename = config_filename or self.trainer_config_file
        subfolder = subfolder or self.trainer_subfolder
        dataset_config_file = dataset_config_file or self.dataset_config_file

        self.config.save(path, filename=config_filename, subfolder=subfolder)
        self.model.save(path, filename=model_filename, config_filename=model_config_filename, save_config=True)
        self.train_dataset.config.save(path, filename=dataset_config_file, subfolder=subfolder)

    def push_to_hub(
        self,
        repo_id: str,
        config_filename: str = None,
        push_model: bool = True,
        model_filename: str = None,
        model_config_filename: str = None,
        subfolder: str = None,
        dataset_config_filename: str = None,
        commit_message: str = None,
        private: bool = False,
    ):
        """
        Push everything to the Hub

        Args:
            repo_id: Path to hub
            config_filename: Trainer config file name
            push_model: Whether to push the model or not
            model_filename: Model file name
            model_config_filename: Model config file name
            subfolder: Path to Trainer files
            dataset_config_filename: Dataset config file name
            commit_message: Commit message for the push
            private: Whether to create a private repo if it doesn't exist already
        """
        config_filename = config_filename or self.trainer_config_file
        subfolder = subfolder or self.trainer_subfolder
        dataset_config_file = dataset_config_filename or self.dataset_config_file

        # create remote repo
        create_repo(repo_id, repo_type="model", exist_ok=True, private=private)

        if not commit_message:
            commit_message = "Hezar: Upload training files"

        # upload train files
        self.config.push_to_hub(
            repo_id,
            filename=config_filename,
            subfolder=subfolder,
            private=private,
            commit_message=commit_message,
        )
        self.train_dataset.config.push_to_hub(
            repo_id,
            filename=dataset_config_file,
            subfolder=subfolder,
            private=private,
            commit_message=commit_message
        )

        # upload model
        if push_model:
            self.model.push_to_hub(
                repo_id,
                filename=model_filename,
                config_filename=model_config_filename,
                commit_message=commit_message,
                private=private,
            )



from .tokenizer import Tokenizer, TokenizerConfig
from .wordpiece import WordPieceTokenizer, WordPieceConfig
from .bpe import BPETokenizer, BPEConfig
from .sentencepiece_bpe import SentencePieceBPETokenizer, SentencePieceBPEConfig



import os
from dataclasses import dataclass
from typing import List, Literal, Union

from gensim.models import fasttext
from huggingface_hub import hf_hub_download

from ..constants import HEZAR_CACHE_DIR
from ..registry import register_embedding
from .embedding import Embedding, EmbeddingConfig


@dataclass
class FastTextConfig(EmbeddingConfig):
    name: str = "fasttext"
    dataset_path: str = None
    vector_size: int = 300
    window: int = 5
    alpha: float = 0.025
    min_count: int = 1
    seed: int = 1
    workers: int = 3
    min_alpha: float = 0.0001
    train_algorithm: Literal["skipgram", "cbow"] = "skipgram"
    cbow_mean = 1
    epochs = 5
    pretrained_path: str = None


@register_embedding("fasttext", config_class=FastTextConfig)
class FastText(Embedding):
    filename = "embedding.model"
    array_filename = "embedding.model.wv.vectors_ngrams.npy"

    def __init__(self, config: FastTextConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.model = self.build()

    def build(self):
        pretrained_path = self.config.get("pretrained_path")
        if pretrained_path:
            if not os.path.isdir(pretrained_path):
                embedding_path = hf_hub_download(
                    pretrained_path,
                    filename=self.filename,
                    subfolder=self.subfolder,
                    cache_dir=HEZAR_CACHE_DIR,
                )

            else:
                embedding_path = os.path.join(
                    pretrained_path,
                    self.subfolder,
                    self.filename,
                )
            embedding_model = fasttext.FastText.load(embedding_path)
        else:
            embedding_model = fasttext.FastText(
                vector_size=self.config.vector_size,
                window=self.config.window,
                sg=1 if self.config.train_algorithm == "skipgram" else 0,
                workers=self.config.workers,
                alpha=self.config.alpha,
                min_alpha=self.config.min_alpha,
                min_count=self.config.min_count,
            )

        return embedding_model

    def train(
        self,
        dataset: List[str],
        epochs: int = 5,
    ):
        self.model.build_vocab(dataset)
        self.model.train(
            dataset,
            epochs=epochs,
            total_examples=self.model.corpus_count,
            total_words=self.model.corpus_total_words,
        )

    def save(
        self,
        path: Union[str, os.PathLike],
        filename: str = None,
        save_config: bool = True,
        config_filename: str = None,
    ):
        filename = filename or self.filename
        config_filename = config_filename or self.config_filename

        os.makedirs(path, exist_ok=True)
        self.config.save(path, config_filename)

        self.model.save(os.path.join(path, filename))

    @property
    def word_vectors(self):
        return self.model.wv

    @property
    def vectors(self):
        return self.model.wv.vectors

    @property
    def vocab(self):
        return self.model.wv.key_to_index



import os
from dataclasses import dataclass, field
from typing import List, Union

from huggingface_hub import create_repo
from tokenizers import normalizers

from ..builders import build_preprocessor
from ..configs import PreprocessorConfig
from ..constants import DEFAULT_NORMALIZER_CONFIG_FILE, DEFAULT_PREPROCESSOR_SUBFOLDER
from ..registry import register_preprocessor
from ..utils import get_logger
from .preprocessor import Preprocessor


logger = get_logger(__name__)


@dataclass
class NormalizerConfig(PreprocessorConfig):
    name: str = field(default="normalizer")
    replace_pattern: str = None
    nfkd: bool = True
    nfkc: bool = True


@register_preprocessor("normalizer", config_class=NormalizerConfig)
class Normalizer(Preprocessor):
    """
    A simple configurable text normalizer
    """
    preprocessor_subfolder = DEFAULT_PREPROCESSOR_SUBFOLDER
    normalizer_config_file = DEFAULT_NORMALIZER_CONFIG_FILE

    def __init__(self, config: NormalizerConfig, **kwargs):
        super().__init__(config, **kwargs)

    def __call__(
        self,
        inputs: Union[str, List[str]],
        replace_pattern: str = None,
        nfkd: bool = None,
        nfkc: bool = None,
        **kwargs,
    ):
        if isinstance(inputs, str):
            inputs = [inputs]

        nfkd = nfkd or self.config.nfkd
        nfkc = nfkc or self.config.nfkc

        if nfkd:
            inputs = [normalizers.NFKD().normalize_str(x) for x in inputs]
        if nfkc:
            inputs = [normalizers.NFKC().normalize_str(x) for x in inputs]

        # TODO add support for other normalizations

        return inputs

    @classmethod
    def load(
        cls,
        hub_or_local_path,
        subfolder=None,
        config_filename=None,
        **kwargs
    ):
        config_filename = config_filename or cls.normalizer_config_file
        subfolder = subfolder or cls.preprocessor_subfolder
        config = NormalizerConfig.load(
            hub_or_local_path,
            filename=config_filename,
            subfolder=subfolder,
        )
        normalizer = build_preprocessor(config.name, config, **kwargs)
        return normalizer

    def push_to_hub(
        self,
        repo_id,
        commit_message: str = None,
        subfolder: str = None,
        config_filename: str = None,
        private: bool = None,
    ):
        """
        Push normalizer config and other optional files to the Hub.

        Args:
            repo_id: Repo id on the Hub
            commit_message: Commit message
            subfolder: Optional subfolder for the normalizer
            config_filename: Optional normalizer config filename
            private: Whether to create a private repo if it does not exist already
        """
        subfolder = subfolder or self.preprocessor_subfolder
        config_filename = config_filename or self.normalizer_config_file

        # create remote repo
        create_repo(repo_id, exist_ok=True, private=private)

        if commit_message is None:
            commit_message = "Hezar: Upload tokenizer and config"

        # upload config
        self.config.push_to_hub(
            repo_id=repo_id,
            filename=config_filename,
            subfolder=subfolder,
            commit_message=commit_message,
        )

    def save(
        self,
        path,
        config_filename=None,
        **kwargs,
    ):
        config_filename = config_filename or self.normalizer_config_file
        os.makedirs(path, exist_ok=True)
        self.config.save(path, filename=config_filename, subfolder=self.preprocessor_subfolder)



"""
A RoBERTa Language Model (HuggingFace Transformers) wrapped by a Hezar Model class
"""
from typing import List, Union

from transformers import RobertaConfig, RobertaModel

from ....models import Model
from ....registry import register_model
from .roberta_lm_config import RobertaLMConfig


@register_model("roberta_lm", config_class=RobertaLMConfig)
class RobertaLM(Model):
    def __init__(self, config, **kwargs):
        super().__init__(config=config, **kwargs)
        self.roberta = RobertaModel(RobertaConfig(**self.config))

    def forward(self, inputs, **kwargs):
        input_ids = inputs.get("token_ids")
        attention_mask = inputs.get("attention_mask", None)
        outputs = self.roberta(
            input_ids=input_ids,
            attention_mask=attention_mask,
            **kwargs,
        )
        return outputs

    def preprocess(self, inputs: Union[str, List[str]], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]
        if "normalizer" in self.preprocessor:
            normalizer = self.preprocessor["normalizer"]
            inputs = normalizer(inputs)
        tokenizer = self.preprocessor["bpe_tokenizer"]
        inputs = tokenizer(inputs, return_tensors="pt", device=self.device)
        return inputs

    def post_process(self, inputs, **kwargs):
        hidden_states = inputs.get("hidden_states", None)
        attentions = inputs.get("attentions", None)
        outputs = {
            "last_hidden_state": inputs.get("last_hidden_state"),
            "hidden_states": hidden_states,
            "attentions": attentions,
        }
        return outputs



from .dataset import Dataset
from .text_classification_dataset import TextClassificationDataset, TextClassificationDatasetConfig
from .sequence_labeling_dataset import SequenceLabelingDataset, SequenceLabelingDatasetConfig



from dataclasses import dataclass
from typing import Dict, List, Tuple

from datasets import load_dataset

from ...configs import DatasetConfig
from ...constants import TaskType
from ...preprocessors import Tokenizer
from ...registry import register_dataset
from ...utils import get_logger
from ..data_collators import SequenceLabelingDataCollator
from .dataset import Dataset


logger = get_logger(__name__)


@dataclass
class SequenceLabelingDatasetConfig(DatasetConfig):
    name: str = "sequence_labeling"
    task: str = TaskType.SEQUENCE_LABELING
    path: str = None
    normalizers: List[Tuple[str, Dict]] = None
    tokenizer_path: str = None
    tags_field: str = None
    tokens_field: str = None
    max_length: int = None
    ignore_index: int = -100
    label_all_tokens: bool = True


@register_dataset("sequence_labeling", config_class=SequenceLabelingDatasetConfig)
class SequenceLabelingDataset(Dataset):
    """
    A sequence labeling dataset class.
    As of now this class is intended for datasets existing on the Hub!

    Args:
        config: Dataset config object
        split: Which split to use
        **kwargs: Extra config parameters to assign to the original config
    """

    def __init__(self, config: SequenceLabelingDatasetConfig, split=None, **kwargs):
        super().__init__(config, **kwargs)
        self.dataset = self._load(split)
        self._extract_labels()
        self.tokenizer = self._build_tokenizer()
        if self.tokenizer:
            self.data_collator = SequenceLabelingDataCollator(self.tokenizer)

    def _load(self, split):
        """
        Load the dataset

        Args:
            split: Dataset split

        Returns:
            The whole dataset
        """
        # TODO: In case we want to make this class work on other types like csv, json, etc. we have to do it here.
        dataset = load_dataset(self.config.path, split=split)
        return dataset

    def _build_tokenizer(self):
        if self.config.tokenizer_path:
            tokenizer = Tokenizer.load(self.config.tokenizer_path)
        else:
            logger.warning("This dataset requires a tokenizer to work. Provide it in config as `tokenizer_path` "
                           "or set it manually as `dataset.tokenizer = your_tokenizer` after building the dataset.")
            tokenizer = None
        return tokenizer

    def _extract_labels(self):
        """
        Extract label names, ids and build dictionaries
        """
        tags_list = self.dataset.features[self.config.tags_field].feature.names
        self.id2label = self.config.id2label = {k: str(v) for k, v in dict(list(enumerate(tags_list))).items()}
        self.label2id = self.config.label2id = {v: k for k, v in self.id2label.items()}
        self.num_labels = self.config.num_labels = len(tags_list)

    def __len__(self):
        return len(self.dataset)

    def _tokenize_and_align(self, tokens, labels):
        tokenized_inputs = self.tokenizer(
            tokens,
            is_split_into_words=True,
            return_word_ids=True,
            padding=True,
            truncation=True,
        )
        word_ids = tokenized_inputs["word_ids"]

        previous_word_idx = None
        label_ids = []
        for word_idx in word_ids:
            # Special tokens have a word id that is None. We set the label to -100, so they are automatically
            # ignored in the loss function.
            if word_idx is None:
                label_ids.append(self.config.ignore_index)
            # We set the label for the first token of each word.
            elif word_idx != previous_word_idx:
                label_ids.append(labels[word_idx])
            # For the other tokens in a word, we set the label to either the current label or -100, depending on
            # the label_all_tokens flag.
            else:
                label_ids.append(labels[word_idx] if self.config.label_all_tokens else self.config.ignore_index)
            previous_word_idx = word_idx

        tokenized_inputs["labels"] = label_ids
        return tokenized_inputs

    def __getitem__(self, index):
        """
        Tokenize inputs and return a dict containing ids, masks, labels, etc.

        Args:
            index: Sample index

        Returns:
            A dict of tokenized text data and labels and some extra stuff
        """
        tokens, tags = self.dataset[index].values()
        inputs = self._tokenize_and_align(tokens, tags)
        return inputs
